"""Tests for dynamic module functionality."""

from tests.data import A, AddDepOnConfigure
from tests.module_utils import create_basic_module, create_dynamic_module
from waku import WakuFactory
from waku.di import scoped


def test_dynamic_module_configuration() -> None:
    """Dynamic modules should properly handle configuration extensions."""
    SomeDynamicModule = create_dynamic_module(
        extensions=[AddDepOnConfigure(scoped(A))],
        name='SomeDynamicModule',
    )
    AppModule = create_basic_module(
        imports=[SomeDynamicModule],
        name='AppModule',
    )
    application = WakuFactory(AppModule).create()

    assert len(application.registry.get(SomeDynamicModule).providers) == 1


def test_module_configuration_with_multiple_import_paths() -> None:
    """Module configuration should be applied only once when imported through multiple paths."""
    SomeModule = create_dynamic_module(
        extensions=[AddDepOnConfigure(scoped(A))],
        name='SomeModule',
    )
    ModuleA = create_basic_module(
        imports=[SomeModule],
        name='ModuleA',
    )
    ModuleB = create_basic_module(
        imports=[SomeModule],
        name='ModuleB',
    )
    AppModule = create_basic_module(
        imports=[ModuleA, ModuleB],
        name='AppModule',
    )
    application = WakuFactory(AppModule).create()

    assert len(application.registry.get(SomeModule).providers) == 1



"""Tests for dependency accessibility validation."""

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Protocol, cast

import pytest
from dishka.exceptions import ImplicitOverrideDetectedError

from tests.data import A, AAliasType, B, C, DependentService, Service, X, Y, Z
from tests.module_utils import create_basic_module
from waku import WakuApplication, WakuFactory
from waku.di import AnyOf, Provider, Scope, contextual, provide, scoped, singleton
from waku.ext.validation import ValidationExtension, ValidationRule
from waku.ext.validation.rules import DependenciesAccessibleRule, DependencyInaccessibleError
from waku.modules import ModuleType


def _impl() -> int:  # pragma: no cover
    return 1


@pytest.fixture
def rule() -> ValidationRule:
    return DependenciesAccessibleRule()


class ApplicationFactoryFunc(Protocol):
    def __call__(
        self,
        root_module: ModuleType,
        *,
        strict: bool = True,
        context: dict[type, object] | None = None,
    ) -> WakuApplication: ...


@pytest.fixture
def application_factory(rule: ValidationRule) -> ApplicationFactoryFunc:
    """Create a factory function for creating WakuApplication instances with validation."""

    def factory(
        root_module: ModuleType,
        *,
        strict: bool = True,
        context: dict[type, object] | None = None,
    ) -> WakuApplication:
        return WakuFactory(
            root_module,
            context=context or {},
            extensions=[ValidationExtension([rule], strict=strict)],
        ).create()

    return factory


@pytest.mark.parametrize(
    ('imports', 'exports', 'should_fail'),
    [
        (False, False, True),
        (False, True, True),
        (True, False, True),
        (True, True, False),
    ],
)
async def test_accessibility_import_export_matrix(
    imports: bool,
    exports: bool,
    should_fail: bool,
    application_factory: ApplicationFactoryFunc,
) -> None:
    """Test all combinations of import/export for dependency accessibility."""
    AModule = create_basic_module(
        providers=[scoped(A)],
        exports=[A] if exports else [],
        name='AModule',
    )
    BModule = create_basic_module(
        providers=[scoped(B)],
        imports=[AModule] if imports else [],
        name='BModule',
    )
    AppModule = create_basic_module(
        imports=[AModule, BModule],
        name='AppModule',
    )

    application = application_factory(AppModule)

    if should_fail:
        with pytest.raises(ExceptionGroup) as exc_info:
            await application.initialize()
        b_module = application.registry.get(BModule)

        error = cast(DependencyInaccessibleError, exc_info.value.exceptions[0].exceptions[0])

        expected_error = DependencyInaccessibleError(A, B, b_module)
        assert isinstance(error, type(expected_error))
        assert str(error) == str(expected_error)
    else:
        await application.initialize()


async def test_accessible_with_exported_and_imported(application_factory: ApplicationFactoryFunc) -> None:
    """Test that dependency is accessible when exported and imported."""
    AModule = create_basic_module(
        providers=[scoped(A), scoped(_impl, provided_type=C)],
        exports=[A, C],
        name='AModule',
    )
    BModule = create_basic_module(
        providers=[scoped(B)],
        imports=[AModule],
        name='BModule',
    )
    AppModule = create_basic_module(
        imports=[AModule, BModule],
        name='AppModule',
    )

    await application_factory(AppModule).initialize()


async def test_accessible_with_global_provider(application_factory: ApplicationFactoryFunc) -> None:
    """Test that global providers are accessible everywhere."""
    AModule = create_basic_module(
        providers=[scoped(A)],
        exports=[A],
        name='AModule',
        is_global=True,
    )
    BModule = create_basic_module(
        providers=[scoped(B)],
        imports=[AModule],
        name='BModule',
    )
    AppModule = create_basic_module(
        imports=[AModule, BModule],
        name='AppModule',
    )

    await application_factory(AppModule).initialize()


@pytest.mark.parametrize('scope', [Scope.APP, Scope.REQUEST])
async def test_accessible_with_contextual_provider(
    rule: ValidationRule,
    scope: Scope,
) -> None:
    """Test that contextual providers are accessible if present in context."""
    Module = create_basic_module(
        providers=[
            contextual(A, scope=scope),
            scoped(B),
        ],
        exports=[B],
        name='Module',
    )
    AppModule = create_basic_module(
        imports=[Module],
        name='AppModule',
    )

    application = WakuFactory(
        AppModule,
        context={A: A()},
        extensions=[ValidationExtension([rule])],
    ).create()
    await application.initialize()


async def test_accessible_with_application_providers(application_factory: ApplicationFactoryFunc) -> None:
    """Test that providers in the application module are accessible."""
    BModule = create_basic_module(
        providers=[scoped(B)],
        exports=[B],
        name='BModule',
    )
    AppModule = create_basic_module(
        providers=[scoped(A)],
        imports=[BModule],
        exports=[A],
        name='AppModule',
    )

    application: WakuApplication = application_factory(AppModule)
    await application.initialize()


async def test_intra_module_access(application_factory: ApplicationFactoryFunc) -> None:
    """Test that providers can access each other within the same module without export."""
    Module = create_basic_module(
        providers=[scoped(A), scoped(B)],
        name='Module',
    )
    AppModule = create_basic_module(
        imports=[Module],
        name='AppModule',
    )

    await application_factory(AppModule).initialize()


async def test_multiple_missing_dependencies(application_factory: ApplicationFactoryFunc) -> None:
    """Test that multiple missing dependencies are all reported."""
    XYModule = create_basic_module(
        providers=[scoped(X), scoped(Y)],
        name='XYModule',
    )
    ZModule = create_basic_module(
        providers=[scoped(Z)],
        name='ZModule',
    )
    AppModule = create_basic_module(
        imports=[XYModule, ZModule],
        name='AppModule',
    )

    with pytest.raises(ExceptionGroup) as exc_info:
        await application_factory(AppModule).initialize()

    errors = cast(list[DependencyInaccessibleError], exc_info.value.exceptions[0].exceptions)
    assert errors[0].required_type is X
    assert errors[1].required_type is Y


async def test_warning_mode(application_factory: ApplicationFactoryFunc) -> None:
    """Test that non-strict mode emits warnings instead of raising."""
    AModule = create_basic_module(
        providers=[scoped(A)],
        name='AModule',
    )
    BModule = create_basic_module(
        providers=[scoped(B)],
        name='BModule',
    )
    AppModule = create_basic_module(
        imports=[AModule, BModule],
        name='AppModule',
    )

    application = application_factory(AppModule, strict=False)

    b_module = application.registry.get(BModule)
    expected_error = DependencyInaccessibleError(A, B, b_module)

    with pytest.warns(Warning, match=re.escape(str(expected_error))):
        await application.initialize()


async def test_any_of_provider(application_factory: ApplicationFactoryFunc) -> None:
    """Test that AnyOf provider works."""

    class AProvider(Provider):
        @provide(scope=Scope.REQUEST)
        def provide_a(self) -> AnyOf[A, AAliasType]:  # noqa: PLR6301
            return A()

    @dataclass()
    class DependsOnAlias:
        a: AAliasType

    AModule = create_basic_module(
        providers=[AProvider()],
        exports=[A, AAliasType],
        name='AModule',
        is_global=True,
    )
    BModule = create_basic_module(
        providers=[scoped(DependsOnAlias)],
        name='BModule',
    )
    AppModule = create_basic_module(
        imports=[AModule, BModule],
        name='AppModule',
    )

    application = application_factory(AppModule)

    async with application, application.container() as request_container:
        a = await request_container.get(A)
        a_alias = await request_container.get(AAliasType)
        assert a is a_alias


async def test_module_cannot_reexport_imported_types(application_factory: ApplicationFactoryFunc) -> None:
    """Test that modules cannot re-export types they don't directly provide.

    This test verifies that a module cannot re-export types that it imports from other modules.
    This enforces proper module encapsulation by preventing transitive dependency access.

    The test should fail with a DependencyInaccessibleError when a module tries to:
        1. Import a type from another module
        2. Re-export that same type
        3. Have other modules depend on that re-exported type
    """
    SharedModule = create_basic_module(
        providers=[scoped(A)],
        exports=[A],
        name='SharedModule',
    )
    ReexportModule = create_basic_module(
        imports=[SharedModule],
        exports=[A],
        name='ReexportModule',
    )
    ConsumerModule = create_basic_module(
        providers=[scoped(B)],
        imports=[ReexportModule],
        name='ConsumerModule',
    )
    AppModule = create_basic_module(
        imports=[ConsumerModule],
        name='AppModule',
    )

    application = application_factory(AppModule)
    with pytest.raises(ExceptionGroup) as exc_info:
        await application.initialize()

    error = cast(DependencyInaccessibleError, exc_info.value.exceptions[0].exceptions[0])
    expected_error = DependencyInaccessibleError(
        required_type=A,
        required_by=B,
        from_module=application.registry.get(ConsumerModule),
    )
    assert str(error) == str(expected_error)


async def test_reexported_module_dependencies(application_factory: ApplicationFactoryFunc) -> None:
    """Test that re-exported module dependencies are accessible in the consumer module."""
    SharedModule = create_basic_module(
        providers=[scoped(A)],
        exports=[A],
        name='SharedModule',
    )
    ReexportModule = create_basic_module(
        providers=[],
        imports=[SharedModule],
        exports=[SharedModule],
        name='ReexportModule',
    )
    # B depends on A, which is re-exported from SharedModule
    ConsumerModule = create_basic_module(
        providers=[scoped(B)],
        imports=[ReexportModule],
        name='ConsumerModule',
    )
    AppModule = create_basic_module(
        imports=[ConsumerModule],
        name='AppModule',
    )

    await application_factory(AppModule).initialize()


async def test_hierarchical_dependencies(application_factory: ApplicationFactoryFunc) -> None:
    """Test deep hierarchical dependencies between modules."""

    @dataclass
    class ServiceA:
        pass

    @dataclass
    class ServiceB:
        a: ServiceA

    @dataclass
    class ServiceC:
        b: ServiceB

    @dataclass
    class ServiceD:
        c: ServiceC

    ModuleA = create_basic_module(
        providers=[scoped(ServiceA)],
        exports=[ServiceA],
        name='ModuleA',
    )
    ModuleB = create_basic_module(
        providers=[scoped(ServiceB)],
        exports=[ServiceB],
        imports=[ModuleA],
        name='ModuleB',
    )
    ModuleC = create_basic_module(
        providers=[scoped(ServiceC)],
        exports=[ServiceC],
        imports=[ModuleB],
        name='ModuleC',
    )
    ModuleD = create_basic_module(
        providers=[scoped(ServiceD)],
        imports=[ModuleC],
        name='ModuleD',
    )
    AppModule = create_basic_module(
        imports=[ModuleD],
        name='AppModule',
    )

    await application_factory(AppModule).initialize()


async def test_transitive_dependencies_not_accessible(application_factory: ApplicationFactoryFunc) -> None:
    """Test that transitive dependencies are not accessible unless re-exported."""

    @dataclass
    class ServiceA:
        pass

    @dataclass
    class ServiceB:
        a: ServiceA

    @dataclass
    class ServiceC:
        # Depends on A which should not be accessible through transitive import
        a: ServiceA

    ModuleA = create_basic_module(
        providers=[scoped(ServiceA)],
        exports=[ServiceA],
        name='ModuleA',
    )
    ModuleB = create_basic_module(
        providers=[scoped(ServiceB)],
        exports=[ServiceB],
        imports=[ModuleA],
        name='ModuleB',
    )
    ModuleC = create_basic_module(
        providers=[scoped(ServiceC)],
        imports=[ModuleB],
        name='ModuleC',
    )
    AppModule = create_basic_module(
        imports=[ModuleC],
        name='AppModule',
    )

    with pytest.raises(ExceptionGroup) as exc_info:
        await application_factory(AppModule).initialize()

    error = cast(DependencyInaccessibleError, exc_info.value.exceptions[0].exceptions[0])
    assert error.required_type is ServiceA


async def test_multiple_module_providers(application_factory: ApplicationFactoryFunc) -> None:
    """Test when multiple modules provide the same type."""

    @dataclass
    class SharedInterface(ABC):
        @abstractmethod
        def get_name(self) -> str: ...

    @dataclass
    class FirstImplementation(SharedInterface):
        def get_name(self) -> str:  # pragma: no cover # noqa: PLR6301
            return 'first'

    @dataclass
    class SecondImplementation(SharedInterface):
        def get_name(self) -> str:  # pragma: no cover  # noqa: PLR6301
            return 'second'

    @dataclass
    class Consumer:
        dependency: SharedInterface

    FirstModule = create_basic_module(
        providers=[scoped(FirstImplementation, provided_type=SharedInterface)],
        exports=[SharedInterface],
        name='FirstModule',
    )
    SecondModule = create_basic_module(
        providers=[scoped(SecondImplementation, provided_type=SharedInterface)],
        exports=[SharedInterface],
        name='SecondModule',
    )
    AppModule = create_basic_module(
        providers=[scoped(Consumer)],
        imports=[FirstModule, SecondModule],
        name='AppModule',
    )

    with pytest.raises(ImplicitOverrideDetectedError):
        await application_factory(AppModule).initialize()


async def test_dependencies_from_indirect_imports_are_not_accessible(
    application_factory: ApplicationFactoryFunc,
) -> None:
    """Test that modules cannot access dependencies from indirect imports.

    This test verifies that a module can only access dependencies from modules it directly imports.
    Dependencies from modules that are imported by the module's imports (transitive dependencies)
    should not be accessible. This enforces explicit dependency declarations and prevents
    implicit dependency chains.

    Test structure:
    SecondLevelModule -> provides and exports Service
    FirstLevelModule  -> imports SecondLevelModule but doesn't re-export it
    ConsumerModule    -> imports FirstLevelModule and depends on Service

    Expected: DependencyInaccessibleError because ConsumerModule cannot access Service
             through FirstLevelModule's import of SecondLevelModule.
    """
    SecondLevelModule = create_basic_module(
        providers=[scoped(Service)],
        exports=[Service],
        name='SecondLevelModule',
    )
    FirstLevelModule = create_basic_module(
        imports=[SecondLevelModule],
        exports=[],
        name='FirstLevelModule',
    )
    ConsumerModule = create_basic_module(
        providers=[scoped(DependentService)],
        imports=[FirstLevelModule],
        name='ConsumerModule',
    )

    AppModule = create_basic_module(
        imports=[ConsumerModule],
        name='AppModule',
    )

    application = application_factory(AppModule)
    with pytest.raises(ExceptionGroup) as exc_info:
        await application.initialize()

    error = cast(DependencyInaccessibleError, exc_info.value.exceptions[0].exceptions[0])
    expected_error = DependencyInaccessibleError(
        required_type=Service,
        required_by=DependentService,
        from_module=application.registry.get(ConsumerModule),
    )
    assert str(error) == str(expected_error)


async def test_with_realistic_graph(application_factory: ApplicationFactoryFunc) -> None:
    @dataclass
    class AsyncEngine:
        pass

    @dataclass
    class AsyncConnection:
        engine: AsyncEngine

    @dataclass
    class AsyncSession:
        connection: AsyncConnection

    DBModule = create_basic_module(
        providers=[
            singleton(AsyncEngine),
            scoped(AsyncConnection),
            scoped(AsyncSession),
        ],
        exports=[
            AsyncEngine,
            AsyncConnection,
            AsyncSession,
        ],
        name='DBModule',
    )

    InfraModule = create_basic_module(
        imports=[DBModule],
        exports=[DBModule],
        is_global=True,
        name='InfraModule',
    )

    @dataclass
    class Settings:
        pass

    @dataclass
    class UserService:
        session: AsyncSession
        settings: Settings

    UsersModule = create_basic_module(providers=[scoped(UserService)], name='UsersModule')

    AppModule = create_basic_module(
        providers=[contextual(Settings, scope=Scope.APP)],
        imports=[InfraModule, UsersModule],
        name='AppModule',
    )

    application = application_factory(AppModule, context={Settings: Settings()})
    await application.initialize()



"""Test configuration and shared fixtures."""

from typing import Any, cast

import pytest


@pytest.fixture(
    params=[
        pytest.param(('asyncio', {'use_uvloop': True}), id='asyncio+uvloop'),
        pytest.param(('asyncio', {'use_uvloop': False}), id='asyncio'),
        pytest.param(('trio', {'restrict_keyboard_interrupt_to_checkpoints': True}), id='trio'),
    ],
    autouse=True,
)
def anyio_backend(request: pytest.FixtureRequest) -> tuple[str, dict[str, Any]]:
    return cast(tuple[str, dict[str, Any]], request.param)



from tests.data import A, C
from tests.module_utils import create_basic_module
from waku import WakuFactory
from waku.di import scoped, singleton


async def test_module_registration_with_dependencies() -> None:
    """Modules should be properly registered with their dependencies and providers."""
    ModuleA = create_basic_module(
        providers=[scoped(A)],
        exports=[A],
        name='ModuleA',
    )
    ModuleB = create_basic_module(
        providers=[singleton(C)],
        imports=[ModuleA],
        name='ModuleB',
    )
    AppModule = create_basic_module(
        imports=[ModuleA, ModuleB],
        name='AppModule',
    )

    application = WakuFactory(AppModule).create()

    async with application, application.container() as request_container:
        await request_container.get(A)
        await application.container.get(C)



from typing import Protocol

from tests.data import UserService
from tests.module_utils import create_basic_module
from waku import WakuFactory
from waku.di import object_, scoped


async def test_object_provider_instance_identity() -> None:
    """Object provider should maintain instance identity and values."""
    instance = UserService(user_id=1337)

    AppModule = create_basic_module(
        providers=[object_(instance, provided_type=UserService)],
        name='AppModule',
    )

    app = WakuFactory(AppModule).create()

    async with app, app.container() as container:
        result = await container.get(UserService)
        assert result is instance
        assert result.user_id == 1337


async def test_provider_with_base_class_type() -> None:
    """Provider should support providing implementation through base class type."""

    class BaseService(Protocol):
        pass

    class ConcreteService(BaseService):
        pass

    AppModule = create_basic_module(
        providers=[scoped(ConcreteService, provided_type=BaseService)],
        name='AppModule',
    )

    app = WakuFactory(AppModule).create()

    async with app, app.container() as container:
        result = await container.get(BaseService)
        assert isinstance(result, ConcreteService)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from waku.exceptions import WakuError

if TYPE_CHECKING:
    from waku.mediator.contracts.event import Event
    from waku.mediator.contracts.request import Request
    from waku.mediator.events.handler import EventHandlerType
    from waku.mediator.requests.handler import RequestHandlerType

__all__ = [
    'EventHandlerAlreadyRegistered',
    'EventHandlerNotFound',
    'ImproperlyConfiguredError',
    'MediatorError',
    'RequestHandlerAlreadyRegistered',
    'RequestHandlerNotFound',
]


class MediatorError(WakuError):
    """Base exception for all mediator-related errors."""


class ImproperlyConfiguredError(MediatorError):
    """Raised when mediator configuration is invalid."""


class RequestHandlerAlreadyRegistered(MediatorError, KeyError):  # noqa: N818
    """Raised when a request handler is already registered.

    Attributes:
        request_type: The type of request that caused the error.
        handler_type: The type of handler that was already registered.
    """

    request_type: type[Request[Any]]
    handler_type: RequestHandlerType[Any, Any]

    def __init__(
        self,
        msg: str,
        request_type: type[Request[Any]],
        handler_type: RequestHandlerType[Any, Any],
    ) -> None:
        super().__init__(msg)
        self.request_type = request_type
        self.handler_type = handler_type


class RequestHandlerNotFound(MediatorError, TypeError):  # noqa: N818
    """Raised when a request handler is not found.

    Attributes:
        request_type: The type of request that caused the error.
    """

    request_type: type[Request[Any]]

    def __init__(self, msg: str, request_type: type[Request[Any]]) -> None:
        super().__init__(msg)
        self.request_type = request_type


class EventHandlerAlreadyRegistered(MediatorError, KeyError):  # noqa: N818
    """Raised when an event handler is already registered.

    Attributes:
        event_type: The type of event that caused the error.
        handler_type: The type of handler that was already registered.
    """

    event_type: type[Event]
    handler_type: EventHandlerType[Any]

    def __init__(
        self,
        msg: str,
        event_type: type[Event],
        handler_type: EventHandlerType[Any],
    ) -> None:
        super().__init__(msg)
        self.event_type = event_type
        self.handler_type = handler_type


class EventHandlerNotFound(MediatorError, TypeError):  # noqa: N818
    """Raised when an event handler is not found.

    Attributes:
        event_type: The type of event that caused the error.
    """

    event_type: type[Event]

    def __init__(self, msg: str, event_type: type[Event]) -> None:
        super().__init__(msg)
        self.event_type = event_type



"""Tests for dependency injection scopes."""

from tests.data import DependentService, RequestContext, Service, UserService
from tests.module_utils import create_basic_module
from waku import WakuFactory
from waku.di import contextual, scoped, singleton, transient


async def test_provider_scope_behavior() -> None:
    """Different provider scopes should behave according to their lifecycle rules."""
    AppModule = create_basic_module(
        providers=[
            scoped(Service),
            transient(lambda: UserService(user_id=2), provided_type=UserService),
        ],
        name='AppModule',
    )

    app = WakuFactory(AppModule).create()

    async with app, app.container() as container:
        # Test transient - new instance each time
        service1 = await container.get(UserService)
        service2 = await container.get(UserService)
        assert service1 is not service2
        assert service1.user_id == 2
        assert service2.user_id == 2

        # Test scoped - same instance per request
        dep1 = await container.get(Service)
        dep2 = await container.get(Service)
        assert dep1 is dep2


async def test_contextual_provider_instance_resolution() -> None:
    """Contextual provider should resolve instances from provided context."""
    AppModule = create_basic_module(
        providers=[contextual(UserService)],
        name='AppModule',
    )

    app = WakuFactory(AppModule).create()

    async with app:
        service_instance = UserService(user_id=4)
        context = {UserService: service_instance}
        async with app.container(context=context) as container:
            result = await container.get(UserService)
            assert result is service_instance
            assert result.user_id == 4


def _create_user_service(ctx: RequestContext) -> UserService:
    return UserService(ctx.user_id)


async def test_request_context_provider_isolation() -> None:
    """Request context providers should maintain isolation between different requests."""
    AppModule = create_basic_module(
        providers=[
            contextual(RequestContext),
            scoped(_create_user_service, provided_type=UserService),
        ],
        name='AppModule',
    )

    app = WakuFactory(AppModule).create()

    async with app:
        # First request
        context1 = {RequestContext: RequestContext(user_id=1)}
        async with app.container(context=context1) as container1:
            user1 = await container1.get(UserService)
            assert user1.user_id == 1

        # Second request with different context
        context2 = {RequestContext: RequestContext(user_id=2)}
        async with app.container(context=context2) as container2:
            user2 = await container2.get(UserService)
            assert user2.user_id == 2


async def test_injected_dependency_instance_sharing() -> None:
    """Injected dependencies should share the same instance within a scope."""
    AppModule = create_basic_module(
        providers=[
            singleton(lambda: 1, provided_type=int),
            scoped(Service),
            scoped(DependentService),
        ],
        name='AppModule',
    )

    app = WakuFactory(AppModule).create()

    async with app, app.container() as container:
        service = await container.get(Service)
        dep = await container.get(DependentService)
        assert dep.service is service


async def test_request_scope_dependency_injection() -> None:
    """Test that request scoped dependencies are properly injected."""
    ServiceModule = create_basic_module(
        providers=[scoped(Service)],
        exports=[Service],
        name='ServiceModule',
    )

    DependentModule = create_basic_module(
        providers=[scoped(DependentService)],
        imports=[ServiceModule],
        name='DependentModule',
    )

    AppModule = create_basic_module(
        imports=[ServiceModule, DependentModule],
        name='AppModule',
    )

    application = WakuFactory(AppModule).create()

    async with application, application.container(context={int: 42}) as request_container:
        service = await request_container.get(Service)
        dependent = await request_container.get(DependentService)

        assert dependent.service is service


async def test_request_scope_with_context() -> None:
    """Test that request scoped dependencies have access to request context."""
    UserModule = create_basic_module(
        providers=[scoped(UserService), contextual(int)],
        name='ServiceModule',
    )

    AppModule = create_basic_module(
        imports=[UserModule],
        name='AppModule',
    )

    application = WakuFactory(AppModule).create()

    async with application, application.container(context={int: 42}) as request_container:
        user_service = await request_container.get(UserService)
        assert user_service.user_id == 42



from dataclasses import dataclass

from tests.module_utils import create_basic_module
from waku import WakuFactory
from waku.di import Scope, contextual, scoped
from waku.testing import override


@dataclass
class Service:
    val: int

    def method(self) -> int:
        return self.val


class ServiceOverride(Service):
    def method(self) -> int:  # noqa: PLR6301
        return 42


async def test_override_helper() -> None:
    AppModule = create_basic_module(
        providers=[
            contextual(int, scope=Scope.APP),
            scoped(Service),
        ],
        name='AppModule',
    )

    val = 1
    application = WakuFactory(AppModule, context={int: val}).create()

    async with application:
        async with application.container() as request_container:
            original_service = await request_container.get(Service)
            assert isinstance(original_service, Service)
            assert original_service.method() == val

        with override(application.container, scoped(ServiceOverride, provided_type=Service)):
            async with application.container() as request_container:
                overrode_service = await request_container.get(Service)
                assert isinstance(overrode_service, ServiceOverride)
                assert overrode_service.method() == 42



"""Tests for module extensions."""

from tests.data import A, AddDepOnConfigure, OnDestroyExt, OnInitExt
from tests.module_utils import create_basic_module
from waku import WakuFactory
from waku.di import scoped


async def test_module_extensions_initialization_and_destruction_order() -> None:
    """Module extensions should be called in correct order during initialization and destruction."""
    calls: list[tuple[type, type]] = []

    GlobalModule = create_basic_module(
        extensions=[OnInitExt(calls), OnDestroyExt(calls)],
        name='GlobalModule',
        is_global=True,
    )

    DatabaseModule = create_basic_module(
        extensions=[OnInitExt(calls), OnDestroyExt(calls)],
        name='DatabaseModule',
    )

    UsersModule = create_basic_module(
        imports=[DatabaseModule],
        extensions=[OnInitExt(calls), OnDestroyExt(calls)],
        name='UsersModule',
    )

    AuthModule = create_basic_module(
        imports=[UsersModule],
        extensions=[OnInitExt(calls), OnDestroyExt(calls)],
        name='AuthModule',
    )

    AppModule = create_basic_module(
        imports=[GlobalModule, DatabaseModule, UsersModule, AuthModule],
        extensions=[OnInitExt(calls), OnDestroyExt(calls)],
        name='AppModule',
    )

    application = WakuFactory(AppModule).create()

    async with application:
        pass

    assert calls == [
        (GlobalModule, OnInitExt),
        (DatabaseModule, OnInitExt),
        (UsersModule, OnInitExt),
        (AuthModule, OnInitExt),
        (AppModule, OnInitExt),
        (AppModule, OnDestroyExt),
        (AuthModule, OnDestroyExt),
        (UsersModule, OnDestroyExt),
        (DatabaseModule, OnDestroyExt),
        (GlobalModule, OnDestroyExt),
    ]

    excepted_modules_order = [
        GlobalModule,
        DatabaseModule,
        UsersModule,
        AuthModule,
        AppModule,
    ]

    for mod, expected_type in zip(application.registry.modules, excepted_modules_order, strict=True):
        assert mod.target is expected_type


def test_module_configure_extension_idempotency() -> None:
    """Module configuration should be applied only once regardless of multiple factory creations."""
    SomeModule = create_basic_module(
        extensions=[AddDepOnConfigure(scoped(A))],
        name='SomeModule',
    )
    AppModule = create_basic_module(
        imports=[SomeModule],
        name='AppModule',
    )

    WakuFactory(AppModule).create()
    application = WakuFactory(AppModule).create()

    assert len(application.registry.get(SomeModule).providers) == 1



"""Common module patterns used across tests."""

from collections.abc import Sequence
from typing import Any

from waku import module
from waku.extensions import ModuleExtension
from waku.modules import DynamicModule, ModuleType


def create_basic_module(
    *,
    providers: Sequence[Any] | None = None,
    exports: Sequence[Any] | None = None,
    imports: Sequence[Any] | None = None,
    name: str | None = None,
    extensions: Sequence[ModuleExtension] | None = None,
    is_global: bool = False,
) -> ModuleType:
    """Create a basic module with given configuration."""
    cls: ModuleType = module(
        providers=list(providers or []),
        exports=list(exports or []),
        imports=list(imports or []),
        extensions=list(extensions or []),
        is_global=is_global,
    )(type(name or 'BasicModule', (object,), {}))

    return cls


def create_dynamic_module(
    *,
    providers: Sequence[Any] | None = None,
    exports: Sequence[Any] | None = None,
    imports: Sequence[Any] | None = None,
    extensions: Sequence[Any] | None = None,
    name: str | None = None,
) -> DynamicModule:
    """Create a dynamic module with given configuration."""

    @module()
    class DynamicModuleParent:
        @classmethod
        def register(cls) -> DynamicModule:
            return DynamicModule(
                parent_module=cls,
                providers=list(providers or []),
                exports=list(exports or []),
                imports=list(imports or []),
                extensions=list(extensions or []),
            )

    if name:  # pragma: no cover
        DynamicModuleParent.__name__ = name

    return DynamicModuleParent.register()



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Final

from waku.di import Provider

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from uuid import UUID

    from waku.di import BaseProvider
    from waku.extensions import ModuleExtension
    from waku.modules._metadata import DynamicModule, ModuleMetadata, ModuleType


__all__ = ['Module']


class Module:
    def __init__(self, module_type: ModuleType, metadata: ModuleMetadata) -> None:
        self.id: Final[UUID] = metadata.id
        self.target: Final[ModuleType] = module_type

        self.providers: Final[Sequence[BaseProvider]] = metadata.providers
        self.imports: Final[Sequence[ModuleType | DynamicModule]] = metadata.imports
        self.exports: Final[Sequence[type[object] | ModuleType | DynamicModule]] = metadata.exports
        self.extensions: Final[Sequence[ModuleExtension]] = metadata.extensions
        self.is_global: Final[bool] = metadata.is_global

    @property
    def name(self) -> str:
        return self.target.__name__

    @cached_property
    def provider(self) -> Provider:
        cls: type[_ModuleProvider] = type(f'{self.name}Provider', (_ModuleProvider,), {})
        return cls(self.providers)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f'Module[{self.name}]'

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Module):  # pragma: no cover
            return False
        return self.id == other.id


class _ModuleProvider(Provider):
    def __init__(self, providers: Iterable[BaseProvider]) -> None:
        super().__init__()
        for provider in providers:
            self.factories.extend(provider.factories)
            self.aliases.extend(provider.aliases)
            self.decorators.extend(provider.decorators)
            self.context_vars.extend(provider.context_vars)



from collections.abc import Sequence
from dataclasses import dataclass
from itertools import chain
from typing import Any, Self, TypeAlias

from waku.di import Provider, Scope, object_, scoped, transient
from waku.extensions import OnModuleConfigure
from waku.mediator import MiddlewareChain
from waku.mediator.contracts.event import EventT
from waku.mediator.contracts.request import RequestT, ResponseT
from waku.mediator.events.handler import EventHandler, EventHandlerType
from waku.mediator.events.map import EventMap
from waku.mediator.events.publish import EventPublisher, SequentialEventPublisher
from waku.mediator.impl import Mediator
from waku.mediator.interfaces import IMediator, IPublisher, ISender
from waku.mediator.middlewares import Middleware
from waku.mediator.requests.handler import RequestHandler, RequestHandlerType
from waku.mediator.requests.map import RequestMap
from waku.mediator.utils import get_request_response_type
from waku.modules import DynamicModule, ModuleMetadata, module

__all__ = [
    'MediatorConfig',
    'MediatorExtension',
    'MediatorModule',
]


_HandlerProviders: TypeAlias = tuple[Provider, ...]


@dataclass(frozen=True, slots=True, kw_only=True)
class MediatorConfig:
    """Configuration for the Mediator extension.

    This class defines the configuration options for setting up the mediator pattern
    implementation in the application.

    Attributes:
        mediator_implementation_type: The concrete implementation class for the mediator
            interface (IMediator). Defaults to the standard Mediator class.
        event_publisher: The implementation class for publishing events. Defaults to `SequentialEventPublisher`.

        middlewares: A sequence of middleware classes that will be applied to the
            mediator pipeline. Middlewares are executed in the order they are defined.
            Defaults to an empty sequence.

    Example:
        ```python
        config = MediatorConfig(
            mediator_implementation_type=CustomMediator,
            middlewares=[LoggingMiddleware(), ValidationMiddleware()],
        )
        ```
    """

    mediator_implementation_type: type[IMediator] = Mediator
    event_publisher: type[EventPublisher] = SequentialEventPublisher
    middlewares: Sequence[Middleware] = ()


@module()
class MediatorModule:
    @classmethod
    def register(cls, config: MediatorConfig | None = None, /) -> DynamicModule:
        """Application-level module for Mediator setup.

        Args:
            config: Configuration for the Mediator extension.
        """
        config_ = config or MediatorConfig()
        return DynamicModule(
            parent_module=cls,
            providers=list(
                chain(
                    cls._create_mediator_providers(config_),
                    cls._create_middleware_chain_provider(config_),
                ),
            ),
            is_global=True,
        )

    @staticmethod
    def _create_mediator_providers(config: MediatorConfig) -> _HandlerProviders:
        return (
            scoped(config.mediator_implementation_type, provided_type=IMediator),
            scoped(config.mediator_implementation_type, provided_type=ISender),
            scoped(config.mediator_implementation_type, provided_type=IPublisher),
            scoped(config.event_publisher, provided_type=EventPublisher),
        )

    @classmethod
    def _create_middleware_chain_provider(cls, config: MediatorConfig) -> _HandlerProviders:
        return (object_(MiddlewareChain(config.middlewares), provided_type=MiddlewareChain),)


class MediatorExtension(OnModuleConfigure):
    def __init__(self) -> None:
        self._request_map = RequestMap()
        self._event_map = EventMap()

    def bind_request(
        self,
        request_type: type[RequestT],
        handler_type: RequestHandlerType[RequestT, ResponseT],
    ) -> Self:
        self._request_map.bind(request_type, handler_type)
        return self

    def bind_event(
        self,
        event_type: type[EventT],
        handler_types: list[EventHandlerType[EventT]],
    ) -> Self:
        self._event_map.bind(event_type, handler_types)
        return self

    def on_module_configure(self, metadata: ModuleMetadata) -> None:
        metadata.providers.extend(
            chain(
                self._create_request_handler_providers(),
                self._create_event_handler_providers(),
            ),
        )

    def _create_request_handler_providers(self) -> _HandlerProviders:
        return tuple(
            transient(
                handler_type,
                provided_type=RequestHandler[request_type, get_request_response_type(request_type)],  # type: ignore[arg-type, valid-type, misc]
            )
            for request_type, handler_type in self._request_map.registry.items()
        )

    def _create_event_handler_providers(self) -> _HandlerProviders:
        def reg_list(provider: Provider, classes: list[type[Any]], base: type[Any]) -> Provider:
            provider.provide_all(*classes)
            provider.provide(lambda: [], provides=list[base])  # type: ignore[valid-type]  # noqa: PIE807
            for cls in classes:

                @provider.decorate
                def foo(many: list[base], one: cls) -> list[base]:  # type: ignore[valid-type]
                    return [*many, one]

            return provider

        return tuple(
            reg_list(Provider(scope=Scope.REQUEST), handler_types, base=EventHandler[event_type])  # type: ignore[valid-type]
            for event_type, handler_types in self._event_map.registry.items()
        )



from __future__ import annotations

import functools
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable, Iterator, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ParamSpec, Protocol, TypeAlias

from waku.mediator.contracts.request import Request, Response

if TYPE_CHECKING:
    from waku.di import AsyncContainer


__all__ = [
    'BaseMiddleware',
    'HandleType',
    'Middleware',
    'MiddlewareChain',
    'MiddlewareContext',
]


HandleType: TypeAlias = Callable[[Request[Any]], Awaitable[Response | None]]

_P = ParamSpec('_P')


@dataclass(kw_only=True, slots=True, frozen=True)
class MiddlewareContext:
    container: AsyncContainer


class BaseMiddleware(ABC):
    @abstractmethod
    async def __call__(self, request: Request[Any], handle: HandleType) -> Response | None: ...


class _MiddlewareFactory(Protocol[_P]):
    def __call__(self, ctx: MiddlewareContext, /, *args: _P.args, **kwargs: _P.kwargs) -> HandleType: ...


class Middleware:
    def __init__(
        self,
        cls: _MiddlewareFactory[_P],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> None:
        self.cls = cls
        self.args = args
        self.kwargs = kwargs

    def __iter__(self) -> Iterator[Any]:
        as_tuple = (self.cls, self.args, self.kwargs)
        return iter(as_tuple)


class MiddlewareChain:
    def __init__(self, middlewares: Sequence[Middleware]) -> None:
        self._middlewares = middlewares

    def wrap(self, container: AsyncContainer, *, handle: HandleType) -> HandleType:
        original_handle = handle
        for cls, args, kwargs in reversed(self._middlewares):
            middleware = cls(
                MiddlewareContext(container=container),
                *args,
                **kwargs,
            )
            handle = functools.update_wrapper(
                wrapper=functools.partial(middleware.__call__, handle=handle),
                wrapped=original_handle,
            )
        return handle



"""Common test data classes used across tests."""

from dataclasses import dataclass
from typing import NewType

from waku import Module
from waku.di import BaseProvider
from waku.extensions import OnModuleConfigure, OnModuleDestroy, OnModuleInit
from waku.modules import ModuleMetadata


# Simple test services
@dataclass
class Service:
    """A simple service for testing."""


@dataclass
class DependentService:
    """A service that depends on another service."""

    service: Service


@dataclass
class RequestContext:
    """A simple request context for testing."""

    user_id: int


@dataclass
class UserService:
    """A user service for testing."""

    user_id: int


# Simple A, B, C services commonly used in dependency tests
@dataclass
class A:
    """Service A for testing dependencies."""


AAliasType = NewType('AAliasType', A)


@dataclass
class B:
    """Service B for testing dependencies."""

    a: A


@dataclass
class C:
    """Service C for testing dependencies."""


@dataclass
class X:
    """Service X for testing dependencies."""


@dataclass
class Y:
    """Service Y for testing dependencies."""


@dataclass
class Z:
    """Service Z for testing dependencies."""

    x: X
    y: Y


# Common module extensions
class OnInitExt(OnModuleInit):
    """Extension that tracks module initialization."""

    def __init__(self, calls: list[tuple[type, type]]) -> None:
        self.calls = calls

    async def on_module_init(self, module: Module) -> None:
        self.calls.append((module.target, type(self)))


class OnDestroyExt(OnModuleDestroy):
    """Extension that tracks module destruction."""

    def __init__(self, calls: list[tuple[type, type]]) -> None:
        self.calls = calls

    async def on_module_destroy(self, module: Module) -> None:
        self.calls.append((module.target, type(self)))


class AddDepOnConfigure(OnModuleConfigure):
    """Extension that adds a dependency during module configuration."""

    def __init__(self, provider: BaseProvider) -> None:
        self.provider = provider

    def on_module_configure(self, metadata: ModuleMetadata) -> None:
        metadata.providers.append(self.provider)



from waku.ext.validation.rules.dependency_accessible import DependenciesAccessibleRule, DependencyInaccessibleError

__all__ = [
    'DependenciesAccessibleRule',
    'DependencyInaccessibleError',
]



from waku.mediator.contracts.request import Request, Response
from waku.mediator.impl import Mediator
from waku.mediator.interfaces import IMediator, IPublisher, ISender
from waku.mediator.middlewares import MiddlewareChain
from waku.mediator.modules import MediatorConfig, MediatorExtension, MediatorModule
from waku.mediator.requests.handler import RequestHandler
from waku.mediator.requests.map import RequestMap

__all__ = [
    'IMediator',
    'IPublisher',
    'ISender',
    'Mediator',
    'MediatorConfig',
    'MediatorExtension',
    'MediatorModule',
    'MiddlewareChain',
    'Request',
    'RequestHandler',
    'RequestMap',
    'Response',
]



from __future__ import annotations

import functools
import typing

from waku.mediator.contracts.request import ResponseT

if typing.TYPE_CHECKING:
    from waku.mediator.contracts.request import Request

__all__ = ['get_request_response_type']


@functools.cache
def get_request_response_type(request_type: type[Request[ResponseT]]) -> type[ResponseT]:
    return typing.cast(type[ResponseT], typing.get_args(request_type.__orig_bases__[0])[0])  # type: ignore[attr-defined]



"""Tests for application lifecycle management."""

from dataclasses import dataclass

from tests.module_utils import create_basic_module
from waku import WakuFactory


@dataclass
class MockLifespanManager:
    """A mock lifespan manager for testing."""

    entered: bool = False
    exited: bool = False

    async def __aenter__(self) -> None:
        self.entered = True

    async def __aexit__(self, *_: object) -> None:
        self.exited = True


async def test_application_lifespan_manager_execution() -> None:
    """Application should execute lifespan managers in order and handle their lifecycle correctly."""
    manager_1 = MockLifespanManager()
    manager_2 = MockLifespanManager()

    AppModule = create_basic_module(name='AppModule')

    application = WakuFactory(AppModule, lifespan=[manager_1, manager_2]).create()

    async with application:
        assert manager_1.entered
        assert manager_2.entered
        assert not manager_1.exited
        assert not manager_2.exited

    assert manager_1.exited
    assert manager_2.exited  # type: ignore[unreachable]



from collections.abc import Callable
from typing import Any

from dishka import (
    AnyOf,
    AsyncContainer,
    FromComponent,
    FromDishka as Injected,
    Provider,
    Scope,
    WithParents,
    alias,
    from_context,
    provide,
    provide_all,
)
from dishka.provider import BaseProvider

__all__ = [
    'AnyOf',
    'AsyncContainer',
    'BaseProvider',
    'FromComponent',
    'Injected',
    'Provider',
    'Scope',
    'WithParents',
    'alias',
    'contextual',
    'from_context',
    'object_',
    'provide',
    'provide_all',
    'provider',
    'scoped',
    'singleton',
    'transient',
]


def provider(
    source: Callable[..., Any] | type[Any],
    *,
    scope: Scope = Scope.REQUEST,
    provided_type: Any = None,
    cache: bool = True,
) -> Provider:
    """Create a Dishka provider for a callable or type.

    Args:
        source: Callable or type to provide as a dependency.
        scope: Scope of the dependency (default: Scope.REQUEST).
        provided_type: Explicit type to provide (default: inferred).
        cache: Whether to cache the instance in the scope.

    Returns:
        Provider: Configured provider instance.
    """
    provider_ = Provider(scope=scope)
    provider_.provide(source, provides=provided_type, cache=cache)
    return provider_


def singleton(
    source: Callable[..., Any] | type[Any],
    *,
    provided_type: Any = None,
) -> Provider:
    """Create a singleton provider (lifetime: app).

    Args:
        source: Callable or type to provide as a singleton.
        provided_type: Explicit type to provide (default: inferred).

    Returns:
        Provider: Singleton provider instance.
    """
    return provider(source, scope=Scope.APP, provided_type=provided_type)


def scoped(
    source: Callable[..., Any] | type[Any],
    *,
    provided_type: Any = None,
) -> Provider:
    """Create a scoped provider (lifetime: request).

    Args:
        source: Callable or type to provide as a scoped dependency.
        provided_type: Explicit type to provide (default: inferred).

    Returns:
        Provider: Scoped provider instance.
    """
    return provider(source, scope=Scope.REQUEST, provided_type=provided_type)


def transient(
    source: Callable[..., Any] | type[Any],
    *,
    provided_type: Any = None,
) -> Provider:
    """Create a transient provider (new instance per injection).

    Args:
        source: Callable or type to provide as a transient dependency.
        provided_type: Explicit type to provide (default: inferred).

    Returns:
        Provider: Transient provider instance.
    """
    return provider(source, scope=Scope.REQUEST, provided_type=provided_type, cache=False)


def object_(
    source: Any,
    *,
    provided_type: Any = None,
) -> Provider:
    """Provide the exact object passed at creation time as a singleton dependency.

    The provider always returns the same object instance, without instantiation or copying.

    Args:
        source: The object to provide as-is.
        provided_type: Explicit type to provide (default: inferred).

    Returns:
        Provider: Provider that always returns the given object.
    """
    return provider(lambda: source, scope=Scope.APP, provided_type=provided_type, cache=True)


def contextual(
    provided_type: Any,
    *,
    scope: Scope = Scope.REQUEST,
) -> Provider:
    """Provide a dependency from the current context (e.g., app/request).

    Args:
        provided_type: The type to resolve from context.
        scope: Scope of the context variable (default: Scope.REQUEST).

    Returns:
        Provider: Contextual provider instance.
    """
    provider_ = Provider()
    provider_.from_context(provided_type, scope=scope)
    return provider_



from __future__ import annotations

from typing import TYPE_CHECKING

from waku.ext.validation import ValidationExtension
from waku.ext.validation.rules import DependenciesAccessibleRule

if TYPE_CHECKING:
    from collections.abc import Sequence

    from waku.extensions import ApplicationExtension

__all__ = ['DEFAULT_EXTENSIONS']

DEFAULT_EXTENSIONS: Sequence[ApplicationExtension] = (
    ValidationExtension(
        [DependenciesAccessibleRule()],
        strict=True,
    ),
)



from typing import cast

from typing_extensions import override

from waku.di import AsyncContainer
from waku.mediator.contracts.event import Event, EventT
from waku.mediator.contracts.request import Request, ResponseT
from waku.mediator.events.handler import EventHandler
from waku.mediator.events.publish import EventPublisher
from waku.mediator.exceptions import EventHandlerNotFound, RequestHandlerNotFound
from waku.mediator.interfaces import IMediator
from waku.mediator.middlewares import MiddlewareChain
from waku.mediator.requests.handler import RequestHandler
from waku.mediator.utils import get_request_response_type


class Mediator(IMediator):
    """Default mediator implementation."""

    def __init__(
        self,
        container: AsyncContainer,
        middleware_chain: MiddlewareChain,
        event_publisher: EventPublisher,
    ) -> None:
        self._container = container
        self._middleware_chain = middleware_chain
        self._event_publisher = event_publisher

    @override
    async def send(self, request: Request[ResponseT]) -> ResponseT:
        """Send a request through the mediator middleware chain.

        Args:
            request: The request to process

        Returns:
            Response from the handler

        Raises:
            RequestHandlerNotFound: If no handler is registered for the request type
        """
        request_type = type(request)
        handler = await self._resolve_request_handler(request_type)
        return await self._handle_request(handler, request)

    @override
    async def publish(self, event: Event) -> None:
        """Publish an event to all registered handlers.

        Args:
            event: The event to publish

        Raises:
            EventHandlerNotFound: If no handlers are registered for the event type
        """
        event_type = type(event)
        handlers = await self._resolve_event_handlers(event_type)
        await self._event_publisher(handlers, event)

    async def _resolve_request_handler(
        self,
        request_type: type[Request[ResponseT]],
    ) -> RequestHandler[Request[ResponseT], ResponseT]:
        response_type = get_request_response_type(request_type)  # type: ignore[arg-type]
        handler_type = cast(
            type[RequestHandler[Request[ResponseT], ResponseT]],
            RequestHandler[request_type, response_type],  # type: ignore[valid-type]
        )

        try:
            return await self._container.get(handler_type)
        except ValueError as err:
            msg = f'Request handler for {request_type.__name__} request is not registered'
            raise RequestHandlerNotFound(msg, request_type) from err

    async def _resolve_event_handlers(
        self,
        event_type: type[EventT],
    ) -> list[EventHandler[EventT]]:
        handler_type = cast(type[EventHandler[EventT]], EventHandler[event_type])  # type: ignore[valid-type]
        try:
            return await self._container.get(list[handler_type])  # type: ignore[valid-type]
        except ValueError as err:
            msg = f'Event handlers for {event_type.__name__} event is not registered'
            raise EventHandlerNotFound(msg, event_type) from err

    async def _handle_request(
        self,
        handler: RequestHandler[Request[ResponseT], ResponseT],
        request: Request[ResponseT],
    ) -> ResponseT:
        wrapped_handler = self._middleware_chain.wrap(self._container, handle=handler.handle)
        result = await wrapped_handler(request)
        return cast(ResponseT, result)



from pytest_mock import MockerFixture

from tests.module_utils import create_basic_module
from waku import WakuApplication, WakuFactory
from waku.extensions import AfterApplicationInit, OnApplicationInit, OnModuleConfigure, OnModuleInit
from waku.modules import Module, ModuleMetadata


async def test_module_init_extension_lifecycle(mocker: MockerFixture) -> None:
    """Module extensions should be called in correct order during module initialization."""
    on_module_configure_mock = mocker.MagicMock()
    on_module_init_mock = mocker.AsyncMock()

    class ModuleOnConfigureExt(OnModuleConfigure):
        def on_module_configure(self, metadata: ModuleMetadata) -> None:  # noqa: PLR6301
            on_module_configure_mock(metadata)

    class ModuleOnInitExt(OnModuleInit):
        async def on_module_init(self, module: Module) -> None:  # noqa: PLR6301
            await on_module_init_mock(module)

    AppModule = create_basic_module(
        extensions=[
            ModuleOnConfigureExt(),
            ModuleOnInitExt(),
        ],
        name='AppModule',
    )

    application = WakuFactory(AppModule).create()
    await application.initialize()

    assert on_module_configure_mock.call_count == 1
    assert isinstance(on_module_configure_mock.call_args[0][0], ModuleMetadata)
    assert on_module_init_mock.call_count == 1
    assert isinstance(on_module_init_mock.call_args[0][0], Module)


async def test_application_init_extensions_single_call(mocker: MockerFixture) -> None:
    """Application init extensions should be called exactly once even with multiple initializations."""
    on_app_init_mock = mocker.AsyncMock()
    after_app_init_mock = mocker.AsyncMock()

    class AppOnInitExt(OnApplicationInit):
        async def on_app_init(self, app: WakuApplication) -> None:  # noqa: PLR6301
            await on_app_init_mock(app)

    class AppAfterInitExt(AfterApplicationInit):
        async def after_app_init(self, app: WakuApplication) -> None:  # noqa: PLR6301
            await after_app_init_mock(app)

    application = WakuFactory(
        create_basic_module(name='AppModule'),
        extensions=[
            AppOnInitExt(),
            AppAfterInitExt(),
        ],
    ).create()

    # Should be called once for the application initialization
    await application.initialize()
    await application.initialize()

    assert on_app_init_mock.call_count == 1
    assert isinstance(on_app_init_mock.call_args[0][0], WakuApplication)
    assert after_app_init_mock.call_count == 1
